/*
 * Copyright (c) 2020 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package vcl

import "text/template"

const ingTmplSrc = `vcl 4.1;

import std;
import directors;
import re2;
import dynamic;

backend vk8s_notfound {
	# 192.0.2.0/24 reserved for docs & examples (RFC5737).
	.host = "192.0.2.255";
	.port = "80";
}

{{- define "ProbeDef"}}
{{- range $name, $svc := .}}
{{- if $svc.Probe}}
{{with $svc.Probe}}
probe {{probeName $name}} {
{{- if ne .URL ""}}
	.url = "{{.URL}}";
{{- else if .Request}}
	.request =
{{- range $reqLine := .Request}}
		"{{$reqLine}}"
{{- end}}
		;
{{- end}}
{{- if .ExpResponse}}
	.expected_response = {{.ExpResponse}};
{{- end}}
{{- if .Timeout}}
	.timeout = {{.Timeout}};
{{- end}}
{{- if .Interval}}
	.interval = {{.Interval}};
{{- end}}
{{- if .Initial}}
	.initial = {{.Initial}};
{{- end}}
{{- if .Window}}
	.window = {{.Window}};
{{- end}}
{{- if .Threshold}}
	.threshold = {{.Threshold}};
{{- end}}
}
{{- end}}
{{- end}}
{{- end}}
{{- end}}
{{- template "ProbeDef" .IntSvcs -}}
{{- template "ProbeDef" .ExtSvcs}}

{{range $name, $svc := .IntSvcs -}}
{{range $addr := $svc.Addresses -}}
backend {{backendName $svc $addr.IP}} {
	.host = "{{$addr.IP}}";
	.port = "{{$addr.Port}}";
{{- with $svc}}
{{- if .HostHeader}}
	.host_header = "{{.HostHeader}}";
{{- end}}
{{- if .ConnectTimeout}}
	.connect_timeout = {{.ConnectTimeout}};
{{- end}}
{{- if .FirstByteTimeout}}
	.first_byte_timeout = {{.FirstByteTimeout}};
{{- end}}
{{- if .BetweenBytesTimeout}}
	.between_bytes_timeout = {{.BetweenBytesTimeout}};
{{- end}}
{{- if .ProxyHeader}}
	.proxy_header = {{.ProxyHeader}};
{{- end}}
{{- if .MaxConnections}}
	.max_connections = {{.MaxConnections}};
{{- end}}
{{- if .Probe}}
	.probe = {{probeName $name}};
{{- end}}
{{- end}}
}
{{end -}}
{{end}}

sub vcl_init {
{{- if .Rules}}
	new vk8s_hosts = re2.set(anchor=both);
	{{- range $rule := .Rules}}
	vk8s_hosts.add("\Q{{$rule.Host}}\E(:\d+)?");
	{{- end}}
	vk8s_hosts.compile();
{{end}}

{{- range $name, $svc := .IntSvcs}}
	{{- $dirType := dirType $svc}}
	new {{dirName $svc}} = directors.{{$dirType}}();
	{{- range $addr := $svc.Addresses}}
	{{dirName $svc}}.add_backend({{backendName $svc $addr.IP}}
		{{- if eq $dirType "random"}}
		, 1.0
		{{- end}}
		);
	{{- end}}
	{{- if eq $dirType "shard"}}
	{{- if $svc.Director.Warmup}}
	{{dirName $svc}}.set_warmup({{$svc.Director.Warmup}});
	{{- end}}
	{{- if $svc.Director.Rampup}}
	{{dirName $svc}}.set_rampup({{$svc.Director.Rampup}});
	{{- end}}
	{{dirName $svc}}.reconfigure();
	{{- end}}
{{end}}
{{- range $name, $svc := .ExtSvcs}}
	new {{resolverName $svc}} = dynamic.resolver();
	{{resolverName $svc}}.set_resolution_type(STUB);
	new {{dirName $svc}} = dynamic.director(
		ttl_from = dns,
		ttl = {{.DNSRetryDelay}},
		resolver = {{resolverName $svc}}.use()
		{{- with $svc}}
		{{- if .ExternalPort}}
		, port = "{{.ExternalPort}}"
		{{- end}}
		{{- if .HostHeader}}
		, host_header = "{{.HostHeader}}"
		{{- end}}
		{{- if .ConnectTimeout}}
		, connect_timeout = {{.ConnectTimeout}}
		{{- end}}
		{{- if .FirstByteTimeout}}
		, first_byte_timeout = {{.FirstByteTimeout}}
		{{- end}}
		{{- if .BetweenBytesTimeout}}
		, between_bytes_timeout = {{.BetweenBytesTimeout}}
		{{- end}}
		{{- if .DomainUsageTimeout}}
		, domain_usage_timeout = {{.DomainUsageTimeout}}
		{{- end}}
		{{- if .FirstLookupTimeout}}
		, first_lookup_timeout = {{.FirstLookupTimeout}}
		{{- end}}
		{{- if .ProxyHeader}}
		, proxy_header = {{.ProxyHeader}}
		{{- end}}
		{{- if .MaxConnections}}
		, max_connections = {{.MaxConnections}}
		{{- end}}
		{{- if .Probe}}
		, probe = {{probeName $name}}
		{{- end}}
		{{- end}}
	);
	{{- if .ResolverIdleTimeout}}
	{{resolverName $svc}}.set_idle_timeout({{.ResolverIdleTimeout}});
	{{- end }}
	{{- if .ResolverTimeout}}
	{{resolverName $svc}}.set_timeout({{.ResolverTimeout}});
	{{- end }}
	{{- if .MaxDNSQueries}}
	{{resolverName $svc}}.set_limit_outstanding_queries({{.MaxDNSQueries}});
	{{- end }}
	{{- if not .FollowDNSRedirects}}
	{{resolverName $svc}}.set_follow_redirects(REDIRECTS_DO_NOT_FOLLOW);
	{{- end }}
{{end}}
{{- range $rule := .Rules}}
	new {{urlMatcher $rule}} = re2.set(posix_syntax=true, anchor=start);
	{{- range $path, $svc := $rule.PathMap}}
	{{urlMatcher $rule}}.add("{{$path}}",
		{{- if $svc.ExternalName}}
		backend={{dirName $svc}}.backend("{{$svc.ExternalName}}"));
		{{- else}}
				backend={{dirName $svc}}.backend());
		{{- end}}
	{{- end}}
	{{urlMatcher $rule}}.compile();
{{end -}}
}

sub vk8s_set_backend {
	set req.backend_hint = vk8s_notfound;
{{- if .Rules}}
	if (vk8s_hosts.match(req.http.Host)) {
		if (vk8s_hosts.nmatches() != 1) {
			# Fail fast when the match was not unique.
			return (fail);
		}
		if (0 != 0) {
			#
		}
		{{- range $i, $rule := .Rules}}
		elsif (vk8s_hosts.which() == {{plusOne $i}}) {
			if ({{urlMatcher $rule}}.match(req.url)) {
				set req.backend_hint = {{urlMatcher $rule}}.backend(select=FIRST);
			}
		}
		{{- end}}
	}
{{- end}}

	if (req.backend_hint == vk8s_notfound) {
{{- if .DefaultService.Name}}
		set req.backend_hint = {{dirName .DefaultService}}.backend();
{{- else}}
		return (synth(404));
{{- end}}
	}
}

sub vcl_miss {
	call vk8s_set_backend;
}

sub vcl_pass {
	call vk8s_set_backend;
}

sub vcl_pipe {
	call vk8s_set_backend;
}

sub vcl_hit {
	if (obj.ttl < 0s) {
		# Set a backend for a background fetch.
		call vk8s_set_backend;
	}
}
`

const ingTmplName = "vcl"

var ingressTmpl = template.Must(template.New(ingTmplName).
	Funcs(fMap).Parse(ingTmplSrc))
